import logging

from algo.constant import Symbol
from algo.data.reader import get_data_reader
from algo.strategy.common import BaseStrategyParams, Strategy

log = logging.getLogger(__name__)


def simulate_trading(simulation_date: str, symbol: Symbol, strategy: Strategy, params: BaseStrategyParams):
    log.info('start simulation')
    data_reader = get_data_reader()
    data_iterator = data_reader.hour_data(simulation_date, symbol)
    result = strategy(data_iterator, params)
    log.info('simulation result', extra={'simulation-result': result})
    return result



import json
from datetime import datetime
from enum import Enum
from typing import NamedTuple

JsonReadyPricePoint = tuple[str, float, float]


class PricePoint(NamedTuple):
    time: datetime
    ask: float
    bid: float

    def to_json(self) -> str:
        return json.dumps([self.time.isoformat(), self.ask, self.bid])

    def to_json_ready(self) -> JsonReadyPricePoint:
        return (self.time.isoformat(), self.ask, self.bid)

    @classmethod
    def from_json_ready(cls, data: JsonReadyPricePoint):
        return cls(*(datetime.fromisoformat(data[0]), data[1], data[2]))


class OrderType(Enum):
    BUY = 'BUY'
    SELL = 'SELL'


class TradeResult(object):
    # TBD: make it named tuple
    def __init__(self, trade_profit: float, trade_expenses: float) -> None:
        self.trade_profit = trade_profit
        self.trade_expenses = trade_expenses

    @property
    def profit(self):
        return self.trade_profit - self.trade_expenses

    def __str__(self) -> str:
        return f'<TradeResult({self.profit}):\t{self.trade_profit}\t{self.trade_expenses})>'


class CurrentState(object):
    def __init__(self) -> None:
        self._in_market = False
        self._open_price: float = None
        self._order_type: OrderType = None
        self._sl_size: float
        self.is_sl_close: bool = False

    @property
    def in_market(self) -> bool:
        return self._in_market

    @property
    def order_type(self) -> OrderType:
        return self._order_type

    def open(self, price_point: PricePoint, order_type: OrderType, sl_size=None):
        self._in_market = True
        self._order_type = order_type
        self._open_price = price_point.bid
        self._sl_size = sl_size
        self.is_sl_close = False

    def is_sl_time(self, price_point: PricePoint):
        """stop loss is negative .... why the hell not :) """
        return self.current_profit(price_point) < self._sl_size

    def current_profit(self, price_point: PricePoint) -> float:
        profit = self._open_price - price_point.bid
        if self._order_type != OrderType.SELL:
            profit *= -1
        return profit

    def close(self, price_point: PricePoint) -> TradeResult:
        """ close the trade and return profit"""
        self._in_market = False
        return TradeResult(
            self.current_profit(price_point),
            price_point.ask - price_point.bid
        )

    def __str__(self) -> str:
        return f'<State: order: {self.order_type}, open: {self._open_price}>'




from datetime import datetime

import rfc3339

# import iso8601      # for date string -> date object


def format_vars(*args, **kwargs):
    """
    method to shorten dubug ie
    log.info('message', extra=d(1))
    instead of
    log.info('message', extra={'v': 1})
    """
    return {'a': args, 'kwa': kwargs}


d = format_vars


def is_dt_naive(dt_value: datetime) -> bool:
    """ checks that datetime doesn't have tz information """
    return dt_value.tzinfo is None or dt_value.tzinfo.utcoffset(d) is None


def dt_to_rfc3339(dt_value: datetime) -> str:
    """
    reverse: iso8601.parse_date(date_string)
    """
    return rfc3339.rfc3339(dt_value)



import json
import logging
import time
from typing import Callable, List, Union

import redis

from ..config import config
from ..constant import Symbol
from .common import PricePoint, PriceStream

log = logging.getLogger(__name__)

CachableDataSource = Callable[[str, Symbol], PriceStream]


class MemoryCache():
    def __init__(self) -> None:
        self._cache_data = {}

    def has(self, cache_key: str):
        return cache_key in self._cache_data

    def get(self, cache_key: str):
        return self._cache_data[cache_key]

    def set(self, cache_key: str, data):
        self._cache_data[cache_key] = data

    def has_lock(self, cache_key: str):
        return self.has(cache_key)

    def set_lock(self, cache_key: str):
        return self.set(cache_key, 1)

    def clear_lock(self, cache_key: str):
        del self._cache_data[cache_key]


memory_cache = MemoryCache()


class RedisCache:
    def __init__(self) -> None:
        self._redis_connection = None

    def _init_redis_connection(self):
        if self._redis_connection is None:
            self._redis_connection: redis.Redis = redis.Redis.from_url(
                config.redis_url)

    def has_lock(self, cache_key: str):
        """TODO: implement lock with appropriate redis tools despite of workaround"""
        self._init_redis_connection()
        return self._redis_connection.get(cache_key) is not None

    def set_lock(self, cache_key: str):
        self._init_redis_connection()
        return self._redis_connection.set(cache_key, 1) is not None

    def clear_lock(self, cache_key: str):
        self._init_redis_connection()
        return self._redis_connection.delete(cache_key)

    def has(self, cache_key: str):
        self._init_redis_connection()
        return self._redis_connection.get(cache_key) is not None

    def get(self, cache_key: str):
        """
        may be generator. Usage uf ujson may improve performance as this class essenstially does unjsoning
        """
        self._init_redis_connection()
        data = self._redis_connection.get(cache_key)
        all_data = []

        for data_portion in json.loads(data):
            all_data.append(PricePoint.from_json_ready(data_portion))

        return all_data

    def set(self, cache_key: str, data: List[PricePoint]):
        self._init_redis_connection()
        json_ready_data = []
        for data_portion in data:
            json_ready_data.append(data_portion.to_json_ready())
        serialized_data = json.dumps(json_ready_data)
        self._redis_connection.set(cache_key, serialized_data)


redis_cache = RedisCache()


def lock_key(cache_key: str) -> str:
    """lock key format"""
    return f'{cache_key}-lock'


def cached_factory(cache_instance: Union[MemoryCache, RedisCache]):
    def _cached(cache_key: str, data_source: CachableDataSource) -> PriceStream:
        log.info('starting cached')
        lock_key_name = lock_key(cache_key)
        if cache_instance.has_lock(lock_key_name):
            log.info('found lock, waiting for data')
            while cache_instance.has_lock(lock_key_name):
                log.info('lock exists waiting')
                time.sleep(0.1)

        if cache_instance.has(cache_key):
            yield from cache_instance.get(cache_key)
        else:
            log.info('no cache reading data')
            cache_instance.set_lock(lock_key_name)
            # no cache
            # here need to arrange case for the multiprocess concurrency see README :synccache:
            all_data = []
            for data_portion in data_source:
                all_data.append(data_portion)
                yield data_portion
            cache_instance.set(cache_key, all_data)
            cache_instance.clear_lock(lock_key_name)
    return _cached


cached = cached_factory(RedisCache())



import logging

from algo.data.reader import PriceStream
from algo.indicator import MAverage
from algo.trading import CurrentState, OrderType, PricePoint

from .common import BaseStrategyParams, TradingResult


class MAStrategyParams(BaseStrategyParams):
    first_avg_length: int
    second_avg_length: int


log = logging.getLogger(__name__)


def act(data_stream: PriceStream, params: MAStrategyParams) -> TradingResult:
    (mavg1_len, mavg2_len) = params
    mavg1 = MAverage(mavg1_len)
    mavg2 = MAverage(mavg2_len)
    state = CurrentState()
    trades = []
    history_price: PricePoint
    for history_price in data_stream:
        mavg1.tick(history_price.bid)
        mavg2.tick(history_price.bid)
        log.debug('price data %s, mavg1: %s:%s', history_price, mavg1, mavg2)
        expected_direction = None  # kind of expected behaviour

        if not (mavg1.ready() and mavg2.ready()):
            # skipping before ready
            continue

        if mavg1.value > mavg2.value:
            expected_direction = OrderType.BUY
        else:
            expected_direction = OrderType.SELL

        # if state.in_market and state.order_type == expected_direction:
        #     # all good nothing to do here
        #     continue
        # el
        if state.in_market and state.order_type != expected_direction:
            trade = state.close(history_price)
            trades.append(
                trade
            )
            log.info('closing trade %s', trade)
        elif not state.in_market:
            state.open(history_price, expected_direction)
            log.info('open trade %s', state)

    total_profit = sum(tr.profit for tr in trades)
    total_expenses = sum(tr.trade_expenses for tr in trades)
    num_of_trades = len(trades)
    log.info('simulation total',
             extra={
                 'trades_len': num_of_trades,
                 'total_profit': total_profit,
                 'sum_of_expenses': total_expenses
             })
    return TradingResult(total_profit, total_expenses, num_of_trades)



import logging

from influxdb_client.client.write_api import SYNCHRONOUS
from v20.pricing import ClientPrice  # donno when v2 was installed ...

from ..config import config
from .connection import DataConnection

log = logging.getLogger(__name__)


class DataReporter(DataConnection):  # pylint: disable=too-few-public-methods
    def __init__(self, url, token, org):
        super().__init__(url, token, org)
        self._write_api = self._client.write_api(write_options=SYNCHRONOUS)

    def report(self, price: ClientPrice):
        if len(price.bids) == 0 or len(price.asks) == 0:
            log.info('no-price-data')
            log.error('unexpected-price-datapoint', extra=dir(price))
            return
        json_body = [
            {
                "measurement": 'price',
                "tags": {
                    "instrument": price.instrument,
                },
                "time": price.time,
                "fields": {
                    "bid": price.bids[0].price,
                    "ask": price.asks[0].price,
                }
            }
        ]
        self._write_api.write(bucket='tsdata', record=json_body)
        log.info('data sent')


def get_data_reporter() -> DataReporter:
    return DataReporter(
        config.influx_url,
        config.influx_token,
        config.influx_org
    )



#!/usr/bin/python3

import json
import logging.config

from algo.config import config
from algo.constant import Symbol
from algo.data.reader import PriceStream, get_data_reader
from algo.util import d

logging.config.fileConfig(
    './logging.conf',
    # TBD: figure out why lack of the following line disables logging inside module
    disable_existing_loggers=False
)

log = logging.getLogger(__name__)


# def serialize_data_record(data_record):
#     """
#     input is a FluxRecord
#     """
#     return data_record['result']['_measurement']


def read_all(data: PriceStream):
    data_points_counter = 0
    total_length = 0  # in chars
    total_ask = 0.0
    total_bid = 0.0
    for data_point in data:
        # log.info('data-point', extra=d(data_point))
        # log.info('data-point', extra=d(data_point[0].isoformat()))
        total_ask += data_point[1]
        total_bid += data_point[2]

        data_serialized = data_point.to_json()
        total_length += len(data_serialized)
        data_points_counter += 1
    return data_points_counter, total_length, total_ask, total_bid


if __name__ == '__main__':
    log.info('playground home will be here')
    data_reader = get_data_reader()

    data_iterator = data_reader.day_data(
        '2022-06-20 00:00:00+00:00', Symbol.JPY)
    log.info('start-reading')
    data_stat = read_all(data_iterator)
    log.info('data-stat', extra=d(data_stat))
    log.info('first-read-complete')
    data_iterator = data_reader.day_data(
        '2022-06-20 00:00:00+00:00', Symbol.JPY)
    log.info('start-second-reading')
    data_stat = read_all(data_iterator)
    log.info('data-stat', extra=d(data_stat))
    log.info('second-read-complete')

    # strategy = moving_average_sample
    # result = simulate_trading(
    #     '2022-06-07',
    #     Symbol.JPY,
    #     strategy.module,
    #     strategy.args(2, 3)
    # )
    # log.info('simulation-complete', extra={'result': result})



from typing import Callable, NamedTuple

from algo.data.reader import PriceStream

# for now moeny are float... hopefully they will start soaring soon :rofl:
Money = float


class TradingResult(NamedTuple):
    profit: Money
    expenses: Money
    number_of_trades: int


BaseStrategyParams = NamedTuple


Strategy = Callable[[PriceStream, BaseStrategyParams], TradingResult]


class StrategyDescriptor(NamedTuple):
    name: str
    module: Strategy
    args: BaseStrategyParams
    description: str
